import argparse
import json

from gameanalysis import reduction

from egta import outerloop


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'quiesce', help="""Compute equilibrium using the quiesce procedure""",
        description="""Samples profiles from small subgames, expanding subgame
        support by best responses to candidate subgame equilibria. For games
        with a large number of players, a reduction should be specified.""")
    parser.add_argument(
        '--dpr', metavar='<role:count,role:count,...>', help="""Specify a DPR
        reduction.""")
    parser.add_argument(
        '--initial-subgame', metavar='<json-file>',
        type=argparse.FileType('r'), help="""Specify an initial subgame to
        outerloop over the remaining strategies in the game. If this is
        specified the first line in output will be the final subgame after
        outerlooping, and the remaining lines will be the equilibria.""")

    parser.add_argument(
        '--regret-thresh', metavar='<reg>', type=float, default=1e-3,
        help="""Regret threshold for a mixture to be considered an equilibrium.
        (default: %(default)s)""")
    parser.add_argument(
        '--dist-thresh', metavar='<norm>', type=float, default=1e-3,
        help="""Norm threshold for two mixtures to be considered distinct.
        (default: %(default)s)""")
    parser.add_argument(
        '--max-resamples', metavar='<resamples>', type=int, default=10,
        help="""Number of times to resample all profiles in a subgame when
        equilibrium is not identified.  (default: %(default)d)""")
    parser.add_argument(
        '--max-subgame-size', metavar='<support>', type=int, default=3,
        help="""Support size threshold, beyond which subgames are not required
        to be explored.  (default: %(default)d)""")
    parser.add_argument(
        '--num-equilibria', metavar='<num>', type=int, default=1,
        help="""Number of equilibria requested to be found. This is mainly
        useful when game contains known degenerate equilibria, but those
        strategies are still useful as deviating strategies. (default:
        %(default)d)""")
    parser.add_argument(
        '--num-backups', metavar='<num>', type=int, default=1, help="""Number
        of backup subgames to pop at a time, when no equilibria are confirmed
        in initial required set.  When games get to this point they can quiesce
        slowly because this by default pops one at a time. Increasing this
        number can get games like tis to quiesce more quickly, but naturally,
        also schedules more, potentially unnecessary, simulations. (default:
        %(default)d)""")
    parser.add_argument(
        '--dev-by-role', action='store_true', help="""Explore deviations in
        role order instead of all at once. By default, when checking for
        beneficial deviations, all role deviations are scheduled at the same
        time. Setting this will check one role at a time. If a beneficial
        deviation is found, then that subgame is scheduled without exploring
        deviations from the other roles.""")
    return parser


def parse_reduction(red):
    strats = (s.strip().split(':') for s in red.strip().split(','))
    return {r.strip(): int(c) for r, c in strats}


def find_eqa(scheduler, game, serial, args):
    if args.dpr is not None:
        red = reduction.DeviationPreserving(
            game.num_strategies, game.num_players,
            serial.from_role_json(parse_reduction(args.dpr), dtype=int))
    else:
        red = reduction.Identity(game.num_strategies, game.num_players)

    initial_subgame = args.initial_subgame
    if initial_subgame is not None:
        initial_subgame = serial.from_subgame_json(
            json.load(initial_subgame))

    eqa, subgame = outerloop.outer_loop(
        scheduler, game, initial_subgame, red=red,
        regret_thresh=args.regret_thresh, dist_thresh=args.dist_thresh,
        max_resamples=args.max_resamples, subgame_size=args.max_subgame_size,
        num_equilibria=args.num_equilibria, num_backups=args.num_backups,
        devs_by_role=args.dev_by_role)

    if initial_subgame is not None:
        json.dump(serial.to_subgame_json(subgame), args.output)
        args.output.write('\n')
    return eqa



import logging
import queue
import sys
import threading
import time
import traceback

from egtaonline import api
from gameanalysis import rsgame
from gameanalysis import utils as gu

from egta import profsched
from egta import utils as eu


_log = logging.getLogger(__name__)


class EgtaScheduler(profsched.Scheduler):
    """A profile scheduler that schedules through egta online

    Parameters
    ----------
    sim_id : int
        The id of the egtaonline simulator to use.
    basegame : BasgeGame
        The gameanalysis basegame representing the game to schedule.
    serial : GameSerializer
        The gameanalysis serializer that represents how to convery array
        profiles into json profiles.
    simultanious_obs : int
        The number of simultanious observations to schedule at a time.
        Egtaonline will use this when scheduling.
    configuration : {key: value}
        Dictionary of configuration to use with the scheduler. Any fields that
        are omitted will be filled in by the defaults specified in the
        simulator.
    sleep_time : int
        Time in seconds between queries to egtaonline to determine if profiles
        have finished. This should probably be set roughly equal to the time it
        takes for a simulation to run.
    max_scheduled : int
        The maximum number of observations to schedule simultaneously. Keeping
        this low helps prevent starving others of flux cycles.
    obs_memory : int
        The amount of memory in MB to allocate for each simulation.
    obs_time : int
        The amount of time in seconds to give each simulation to run. Too low
        and long running simulations will get cancelled giving you biased
        samples, too long and it will take longer to schedule jobs on flux.
    auth_token : str, optional
        Authorization token for egtaonline. If unspecified, egtaonline will
        search for an appropriate file containing it.
    """

    def __init__(self, sim_id, basegame, serial, simultanious_obs,
                 configuration, sleep_time, max_scheduled, obs_memory,
                 obs_time, auth_token=None):
        self._game = rsgame.basegame_copy(basegame)
        self._serial = serial

        # {hprof: (lock, obs, [id, num_sched, num_receved])}
        self._profiles = {}
        self._prof_lock = threading.Lock()
        # {prof_id: ^above^}
        self._num_running_profiles = 0
        self._prof_ids = {}
        self._runprof_lock = threading.Lock()
        self._pending_profiles = queue.Queue()

        self._sleep_time = sleep_time
        self._max_running = max_scheduled - 1
        self._auth_token = auth_token
        self._sim_id = sim_id
        self._configuration = configuration
        self._obs_memory = obs_memory
        self._obs_time = obs_time
        self._simult_obs = simultanious_obs

        self._api = None
        self._sched = None
        self._running = False
        self._thread_failed = False

    def schedule(self, profile):
        hprof = gu.hash_array(profile)
        with self._prof_lock:
            val = self._profiles.setdefault(
                hprof, (threading.Lock(), queue.Queue(), [None, 0, 0]))

        _, que, _ = val
        if self._num_running_profiles > self._max_running:
            _log.debug("new profile queued %s", profile)
            self._pending_profiles.put((hprof, val))
        else:
            self._schedule(hprof, val)
        return _EgtaPromise(que, self)

    # TODO When scheduling many profiles at once, this will repeatedly remove
    # and readd the profile to the scheduler, making a lot of requests. This
    # shouldn't be that bad as this isn't really a bottleneck, but it could be
    # made more efficient.
    def _schedule(self, hprof, val):
        profile = hprof.array
        lock, que, nums = val
        prof_id = nums[0]
        with lock:
            nums[1] += 1  # num scheduled

            if prof_id is None:
                _log.debug("new profile scheduled %s", profile)
                prof_id = self._sched.add_profile(
                    self._serial.to_prof_str(profile), nums[1]).id
                nums[0] = prof_id
                self._prof_ids[prof_id] = val
                with self._runprof_lock:
                    self._num_running_profiles += 1

            elif nums[2] < nums[1]:
                self._prof_ids[prof_id] = val
                _log.debug("old profile scheduled %s", profile)
                self._sched.remove_profile(prof_id)
                self._sched.add_profile(
                    self._serial.to_prof_str(profile), nums[1])
                with self._runprof_lock:
                    self._num_running_profiles += 1

    def _update_counts(self):
        try:
            while self._running:
                # First update requirements and mark completed
                _log.info("query scheduler %d", self._sched.id)
                reqs = self._sched.get_requirements().scheduling_requirements
                _log.debug("got reqs %s", reqs)
                for req in reqs:
                    prof_id = req['id']
                    if prof_id not in self._prof_ids:
                        continue  # race condition # pragma: no cover
                    lock, que, nums = self._prof_ids[prof_id]

                    # Check if updated counts is greater than before
                    if req['current_count'] <= nums[2]:
                        continue
                    _log.debug("can update profile %d", prof_id)

                    with lock:
                        # If so, get observations from profile, and update data
                        # This uses fact that recent profiles are listed first,
                        # this will not work properly if this assumption is
                        # violated
                        egta_prof = self._api.get_profile(prof_id)

                        # Sometimes egta returns invalid json
                        jobs = None
                        valid = False
                        while not valid:
                            jobs = egta_prof.get_info('observations')
                            valid = all(o['symmetry_groups'] is not None for o
                                        in jobs.observations)
                        _log.debug("obs json %s",  jobs)
                        # Parse all and slice to have accurate counts
                        new_obs = self._serial.from_samplepay_json(jobs)
                        new_obs = new_obs[:new_obs.shape[0] - nums[2]]
                        for obs in new_obs:
                            que.put(obs)
                        finished = max(min(nums[1] - nums[2], len(new_obs)), 0)
                        nums[2] += len(new_obs)
                        with self._runprof_lock:
                            self._num_running_profiles -= finished

                # Now schedule any pending profiles
                while (not self._pending_profiles.empty() and
                       self._num_running_profiles <= self._max_running):
                    hprof, val = self._pending_profiles.get()
                    self._schedule(hprof, val)

                # Now wait some time
                time.sleep(self._sleep_time)

        except Exception as ex:  # pragma: no cover
            self._thread_failed = True
            exc_type, exc_value, exc_traceback = sys.exc_info()
            _log.critical(''.join(traceback.format_exception(
                exc_type, exc_value, exc_traceback)))
            for _, que, [__, sched, rec] in self._profiles.values():
                for _ in range(sched - rec):
                    que.put(None)  # Notify all waiting threads
            raise ex

    def __enter__(self):
        self._thread_failed = False
        self._api = api.EgtaOnlineApi(self._auth_token)
        name = 'egta_' + eu.random_string(20)

        # Create game to get initial profile data
        gamea = None
        try:
            gamea = self._api.create_game(self._sim_id, name,
                                          self._game.num_all_players,
                                          self._configuration)
            _log.debug("created temp game %s %d", name, gamea.id)
            for role, count, strats in zip(self._serial.role_names,
                                           self._game.num_players,
                                           self._serial.strat_names):
                gamea.add_role(role, count)
                for strat in strats:
                    gamea.add_strategy(role, strat)
            profiles = gamea.get_observations().profiles
        finally:
            if gamea is not None:
                gamea.destroy_game()

        # Parse profiles
        for jprof in profiles:
            prof, spays = self._serial.from_profsamplepay_json(jprof)
            hprof = gu.hash_array(prof)
            que = queue.Queue()
            for pay in spays:
                que.put(pay)
            val = (threading.Lock(), que, [jprof.id, 0, len(spays)])
            self._profiles[hprof] = val

        # Create and start scheduler
        self._sched = self._api.create_generic_scheduler(
            self._sim_id, name, True, self._obs_memory,
            self._game.num_all_players, self._obs_time, self._simult_obs, 1,
            self._configuration)
        _log.info("created scheduler %s %d", name, self._sched.id)
        for role, count in zip(self._serial.role_names,
                               self._game.num_players):
            self._sched.add_role(role, count)
        self._running = True
        thread = threading.Thread(target=self._update_counts, daemon=True)
        thread.start()

        return self

    def __exit__(self, *args):
        self._running = False
        if self._sched is not None:
            self._sched.deactivate()
        if self._api is not None:
            self._api.close()
        _log.info("deactivated scheduler %d", self._sched.id)


class _EgtaPromise(profsched.Promise):
    def __init__(self, que, sched):
        self._val = None
        self._queue = que
        self._sched = sched

    def get(self):
        if self._val is None:
            self._val = self._queue.get()
        if self._sched._thread_failed:
            raise RuntimeError("a thread failed, checl log for details")  # pragma: no cover # noqa
        else:
            return self._val



import logging
import queue
import sys
import threading
import traceback

import numpy as np
from gameanalysis import collect
from gameanalysis import nash
from gameanalysis import reduction
from gameanalysis import regret
from gameanalysis import rsgame
from gameanalysis import subgame
from gameanalysis import utils


_log = logging.getLogger(__name__)


def outer_loop(prof_sched, game, initial_subgame=None, *, red=None,
               regret_thresh=1e-3, dist_thresh=1e-3, max_resamples=10,
               subgame_size=3, num_equilibria=1, num_backups=1,
               devs_by_role=False):
    """Outerloop a game using a scheduler

    If initial_subgame is unspecified, this will instead perform only the inner
    loop of the full game specified.

    Parameters
    ----------
    prof_sched : Scheduler
        The scheduler used to generate payoff data for profiles.
    game : BaseGame
        The gameanalysis basegame for the game to find equilibria on.
    initial_subgame : subgame array, optional
        The initial subgame to innerloop before checking for deviations
        outside. If unspecified, the whole game will be used, effectively
        making this one iteration of the inner loop with no outer looping.
    red : Reduction, optional
        The reduction to apply to the game before scheduling. Using a reduction
        will sample fewer profiles, while only approximating equilibria.
    regret_thresh : float, optional
        The maximum regret to consider an equilibrium an equilibrium.
    dist_thresh : float, optional
        The minimum norm between two mixed profiles for them to be considered
        distinct.
    max_resamples : int > 0, optional
        The maximum number of times to resample a subgame when no equilibria
        can be found before giving up.
    subgame_size : int > 0, optional
        The maximum subgame support size with which beneficial deviations must
        be explored. Subgames with support larger than this are queued and only
        explored in the event that no equilibrium can be found in beneficial
        deviations smaller than this.
    num_equilibria : int > 0, optional
        The number of equilibria to attempt to find. Only one is guaranteed,
        but this might be beneifical if the game has a known degenerate
        equilibria, but one which is still helpful as a deviating strategy.
    num_backups : int > 0, optional
        In the event that no equilibrium can be found in beneficial deviations
        to small subgames, other subgames will be explored. This parameter
        indicates how many subgames for each role should be explored.
    devs_by_role : boolean, optional
        If specified, deviations will only be explored for each role in order,
        proceeding to the next role only when no beneficial deviations are
        found. This can reduce the number of profiles sampled, but may also
        fail to find certain equilibria due to the different path through
        subgames.
    """
    if red is None:
        red = reduction.Identity(game.num_strategies, game.num_players)
    assert rsgame.basegame_copy(game) == red.full_game, \
        "reduction must match game format"

    return _OuterLoop(prof_sched, red, regret_thresh=regret_thresh,
                      dist_thresh=dist_thresh, max_resamples=max_resamples,
                      subgame_size=subgame_size, num_equilibria=num_equilibria,
                      num_backups=num_backups,
                      devs_by_role=devs_by_role).run(initial_subgame)


class _OuterLoop(object):
    """Object to keep track of outer loop progress"""

    def __init__(self, prof_sched, reduction, regret_thresh, dist_thresh,
                 max_resamples, subgame_size, num_equilibria, num_backups,
                 devs_by_role):
        # Input data
        self._sched = _Scheduler(prof_sched, reduction)
        self._game = reduction.red_game
        self._full_sched = self._sched.get_subgame_scheduler(
            np.ones(self._game.num_role_strats, bool))
        self._init_role_dev = 0 if devs_by_role else None
        self._num_eqa = num_equilibria
        self._regret_thresh = regret_thresh
        self._params = dict(
            regret_thresh=regret_thresh, dist_thresh=dist_thresh,
            max_resamples=max_resamples, subgame_size=subgame_size,
            num_equilibria=num_equilibria, num_backups=num_backups,
            devs_by_role=devs_by_role)

        # Bookkeeping
        self._equilibria = []
        self._best_responses = []
        self._threads = queue.Queue()
        self._run_lock = threading.Lock()
        self._thread_failed = False

    def _add_deviations(self, mix, role_index):
        _log.info('scheduling outerloop deviations from %s for role %s',
                  mix, role_index)
        thread = threading.Thread(
            target=lambda: self._run_deviations(mix, role_index))
        thread.start()
        self._threads.put(thread)

    def _run_deviations(self, mix, role_index):
        try:
            support = mix > 0
            game = self._full_sched.get_deviations(support, 1, role_index)
            gains = regret.mixture_deviation_gains(game, mix)
            if role_index is None:
                assert not np.isnan(gains).any(), "There were nan regrets"
                if np.all(gains <= self._regret_thresh):  # Found equilibrium
                    self._equilibria.append(mix)  # atomic
                else:
                    for rgains, rs in zip(game.role_split(gains),
                                          game.role_starts):
                        br = np.argmax(rgains)
                        self._best_responses.append(rs + br)  # atomic

            else:  # Set role index
                rgains = game.role_split(gains)[role_index]
                assert not np.isnan(rgains).any(), "There were nan regrets"
                if np.all(rgains <= self._regret_thresh):  # No deviations
                    role_index += 1
                    if role_index < game.num_roles:  # Explore next deviation
                        self._add_deviations(mix, role_index)
                    else:  # found equilibrium
                        self._equilibria.append(mix)  # atomic
                else:
                    br = game.role_starts[role_index] + np.argmax(rgains)
                    self._best_responses.append(br)  # atomic
        except Exception as ex:  # pragma: no cover
            self._thread_failed = True
            exc_type, exc_value, exc_traceback = sys.exc_info()
            _log.critical(''.join(traceback.format_exception(
                exc_type, exc_value, exc_traceback)))
            raise ex

    def run(self, initial_subgame=None):
        last_mask = np.zeros(self._game.num_role_strats, bool)
        mask = (np.ones(self._game.num_role_strats, bool)
                if initial_subgame is None else initial_subgame)

        with self._run_lock:
            assert self._threads.empty()
            self._equilibria.clear()
            self._thread_failed = False

            while (len(self._equilibria) < self._num_eqa
                   and not last_mask.all()):
                # Generate view
                np.copyto(last_mask, mask)
                self._equilibria.clear()
                self._best_responses.clear()
                sub_sched = self._sched.get_subgame_scheduler(mask)
                sub_game = subgame.subgame(self._game, mask)

                # get inner loop equilibria and
                eqa = _InnerLoop(sub_sched, sub_game, **self._params).run()
                for eqm in eqa:
                    self._add_deviations(subgame.translate(eqm, mask),
                                         self._init_role_dev)
                while not self._threads.empty():
                    if self._thread_failed:
                        raise RuntimeError("a thread failed, checl log for details")  # pragma: no cover # noqa
                    else:
                        self._threads.get().join()

                # Add best responses, or random if there were none
                if self._best_responses or mask.all():
                    mask[self._best_responses] = True
                else:
                    # This will happen if there's only one equilibria in the
                    # subgame, and it's a true equilibria
                    mask[np.random.choice(np.nonzero(~mask)[0])] = True

            eqa = (np.concatenate([eq[None] for eq in self._equilibria])
                   if self._equilibria
                   else np.empty((0, self._game.num_role_strats), float))
            return eqa, last_mask


class _InnerLoop(object):
    """Object to keep track of inner loop progress"""

    def __init__(self, sched, game, regret_thresh, dist_thresh, max_resamples,
                 subgame_size, num_equilibria, num_backups, devs_by_role):
        # Data
        self._sched = sched
        self._game = game

        # Parameters
        self._num_eq = num_equilibria
        self._num_backups = num_backups
        self._sub_size = subgame_size
        self._max_resamples = max_resamples
        self._regret_thresh = regret_thresh
        self._dist_thresh = dist_thresh
        self._init_role_dev = 0 if devs_by_role else None

        # Bookkeeping
        self._threads = queue.Queue()
        self._exp_subgames = collect.BitSet()
        self._exp_subgames_lock = threading.Lock()
        self._exp_mix = collect.MixtureSet(dist_thresh)
        self._exp_mix_lock = threading.Lock()
        self._backups = [queue.PriorityQueue() for _
                         in range(self._game.num_roles)]
        self._equilibria = []
        self._run_lock = threading.Lock()  # only one `run` at a time
        self._nash_lock = threading.BoundedSemaphore(3)  # only 3 `mixed_nash` at a time # noqa
        self._thread_failed = False

    def _add_subgame(self, sub_mask, count):
        if count > self._max_resamples:  # pragma: no cover
            _log.warning("couldn't find equilibrium in subgame %s", sub_mask)
            return
        with self._exp_subgames_lock:
            schedule = count > 1 or self._exp_subgames.add(sub_mask)
        if schedule:
            _log.info('scheduling subgame %s', sub_mask)
            thread = threading.Thread(
                target=lambda: self._run_subgame(sub_mask, count))
            thread.start()
            self._threads.put(thread)

    def _run_subgame(self, sub_mask, count):
        try:
            game = subgame.subgame(self._sched.get_subgame(sub_mask, count),
                                   sub_mask)
            with self._nash_lock:
                eqa = subgame.translate(nash.mixed_nash(
                    game, regret_thresh=self._regret_thresh,
                    dist_thresh=self._dist_thresh, processes=1), sub_mask)

            if eqa.size:
                for eqm in eqa:
                    self._add_deviations(eqm, self._init_role_dev)
            else:
                self._add_subgame(sub_mask, count + 1)  # pragma: no cover
        except Exception as ex:  # pragma: no cover
            self._thread_failed = True
            exc_type, exc_value, exc_traceback = sys.exc_info()
            _log.critical(''.join(traceback.format_exception(
                exc_type, exc_value, exc_traceback)))
            raise ex

    def _add_deviations(self, mix, role_index):
        with self._exp_mix_lock:
            unseen = ((role_index is not None and role_index > 0)
                      or self._exp_mix.add(mix))
        if unseen:
            _log.info('scheduling deviations from %s with role %s', mix,
                      role_index)
            thread = threading.Thread(
                target=lambda: self._run_deviations(mix, role_index))
            thread.start()
            self._threads.put(thread)

    def _run_deviations(self, mix, role_index):
        try:
            support = mix > 0
            game = self._sched.get_deviations(support, 1, role_index)
            gains = regret.mixture_deviation_gains(game, mix)
            if role_index is None:
                assert not np.isnan(gains).any(), "There were nan regrets"
                if np.all(gains <= self._regret_thresh):  # Found equilibrium
                    self._equilibria.append(mix)  # atomic
                else:
                    for ri, rgains in enumerate(game.role_split(gains)):
                        self._queue_subgames(support, rgains, ri)

            else:  # Set role index
                rgains = game.role_split(gains)[role_index]
                assert not np.isnan(rgains).any(), "There were nan regrets"
                if np.all(rgains <= self._regret_thresh):  # No deviations
                    role_index += 1
                    if role_index < game.num_roles:  # Explore next deviation
                        self._add_deviations(mix, role_index)
                    else:  # found equilibrium
                        self._equilibria.append(mix)  # atomic
                else:
                    self._queue_subgames(support, rgains, role_index)
        except Exception as ex:  # pragma: no cover
            self._thread_failed = True
            exc_type, exc_value, exc_traceback = sys.exc_info()
            _log.critical(''.join(traceback.format_exception(
                exc_type, exc_value, exc_traceback)))
            raise ex

    def _queue_subgames(self, support, role_gains, role_index):
        rs = self._game.role_starts[role_index]
        back = self._backups[role_index]

        # Handle best response
        br = np.argmax(role_gains)
        if (role_gains[br] > self._regret_thresh
                and support.sum() < self._sub_size):
            br_sub = support.copy()
            br_sub[rs + br] = True
            self._add_subgame(br_sub, 1)
        else:
            br = None  # Add best response to backup

        for si, gain in enumerate(role_gains):
            if si == br:
                continue
            sub = support.copy()
            sub[rs + si] = True
            back.put((-gain, id(sub), sub))  # id for tie-breaking

    def run(self):
        with self._run_lock:
            # Clear Data structures
            assert self._threads.empty()
            self._exp_subgames.clear()
            self._exp_mix.clear()
            for backup in self._backups:
                while not backup.empty():
                    # This only happens when run twice in a row
                    backup.get()  # pragma: no cover
            self._equilibria.clear()
            self._thread_failed = False

            # Quiesce first time
            for sub_mask in self._game.pure_subgames():
                self._add_subgame(sub_mask, 1)
            while not self._threads.empty():
                if self._thread_failed:
                    raise RuntimeError("a thread failed, checl log for details")  # pragma: no cover # noqa
                else:
                    self._threads.get().join()

            # Add backup games and quiesce again if less than desired number of
            # equilibria
            while (len(self._equilibria) < self._num_eq
                   and not all(q.empty() for q in self._backups)):
                for back in self._backups:
                    for _ in range(self._num_backups):
                        if back.empty():
                            break
                        self._add_subgame(back.get()[2], 1)
                while not self._threads.empty():
                    if self._thread_failed:
                        raise RuntimeError("a thread failed, checl log for details")  # pragma: no cover # noqa
                    else:
                        self._threads.get().join()

            # Return equilibria
            if self._equilibria:
                return np.concatenate([eq[None] for eq in self._equilibria])
            else:
                return np.empty((0, self._game.num_role_strats), float)  # pragma: no cover # noqa


class _Scheduler(object):
    """Scheduler abstraction that allows scheduling deviations and subgames as
    primitives"""

    def __init__(self, prof_sched, reduction):
        self._red = reduction
        self._prof_sched = prof_sched
        self._profiles = {}
        self._lock = threading.Lock()

    def _get_game(self, profiles, count):
        promises = []
        with self._lock:
            for prof in profiles:
                hprof = utils.hash_array(prof)
                prof = self._profiles.setdefault(
                    hprof, _PayoffData(self._prof_sched, prof))
                promise = prof.schedule(count)
                promises.append(promise)
        payoffs = np.concatenate([prom.get()[None] for prom in promises])
        return self._red.reduce_game(rsgame.game_copy(
            self._red.full_game, profiles, payoffs), True)

    def _subgame_profiles(self, subgame_mask):
        return self._red.expand_profiles(subgame.translate(
            subgame.subgame(self._red.red_game, subgame_mask).all_profiles(),
            subgame_mask))

    def get_subgame_scheduler(self, subgame_mask):
        num_strats = self._red.full_game.role_reduce(subgame_mask)
        reduction = self._red.__class__(num_strats,
                                        self._red.full_game.num_players,
                                        self._red.red_game.num_players)
        return _SubgameScheduler(self, reduction, subgame_mask)


class _SubgameScheduler(object):
    """Scheduler that schedules subgames"""

    def __init__(self, qsched, red, subgame_mask):
        self._qsched = qsched
        self._red = red
        self._submask = subgame_mask

    def get_subgame(self, subgame_mask, count):
        full_subgame = np.zeros(self._submask.size, bool)
        full_subgame[self._submask] = subgame_mask
        profiles = self._qsched._subgame_profiles(full_subgame)
        return subgame.subgame(self._qsched._get_game(profiles, count),
                               self._submask)

    def get_deviations(self, subgame_mask, count, role_index):
        full_subgame = np.zeros(self._submask.size, bool)
        full_subgame[self._submask] = subgame_mask
        subgame_profiles = self._qsched._subgame_profiles(full_subgame)
        dev_profiles = subgame.translate(
            self._red.expand_deviation_profiles(subgame_mask, role_index),
            self._submask)
        profiles = np.concatenate((subgame_profiles, dev_profiles))
        return subgame.subgame(self._qsched._get_game(profiles, count),
                               self._submask)


class _PayoffData(object):
    """Payoff data for a profile"""

    def __init__(self, prof_sched, profile):
        self._prof_sched = prof_sched
        self._profile = profile
        self._payoffs = np.zeros(profile.size, float)
        self._count = 0
        self._scheduled = 0

        self._lock = threading.Lock()
        self._cond = threading.Condition()
        self._promises = queue.Queue()
        self._thread_failed = False

    def _keep_going(self):
        with self._lock:
            return self._count < self._scheduled

    def _update(self):
        try:
            while self._keep_going():
                pay = self._promises.get().get()
                with self._lock:
                    self._count += 1
                    self._payoffs += (pay - self._payoffs) / self._count
                with self._cond:
                    self._cond.notify_all()
        except Exception as ex:  # pragma: no cover
            self._thread_failed = True
            exc_type, exc_value, exc_traceback = sys.exc_info()
            _log.critical(''.join(traceback.format_exception(
                exc_type, exc_value, exc_traceback)))
            with self._cond:  # We must notify other threads that this failed
                self._cond.notify_all()
            raise ex

    def schedule(self, n):
        with self._lock:
            if self._count == self._scheduled and self._scheduled < n:
                threading.Thread(target=self._update).start()
            while self._scheduled < n:
                self._scheduled += 1
                self._promises.put(self._prof_sched.schedule(self._profile))
        return _PayoffPromise(self, n)


class _PayoffPromise(object):
    def __init__(self, pdata, n):
        self._pdata = pdata
        self._n = n

    def get(self):
        with self._pdata._cond:
            if self._pdata._thread_failed:
                raise RuntimeError("a thread failed, checl log for details")  # pragma: no cover # noqa
            else:
                self._pdata._cond.wait_for(
                    lambda: self._pdata._count >= self._n)
        return self._pdata._payoffs



import argparse
import json
import logging
import sys

from egtaonline import api

from egta import egtasched


_log = logging.getLogger(__name__)


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'egta', help="""Schedule simulations through egta online""",
        description="""Sample profiles from egta online.""")
    parser.add_argument(
        '--conf', '-c', metavar='<conf-json>', type=argparse.FileType('r'),
        help="""The configuration file for the simulator. This only needs to be
        specified if a configuration couldn't be found in the specified
        game.""")
    parser.add_argument(
        '--sim-id', metavar='<sim-id>', type=int, help="""The id of the
        simulator to use. If a game id was specified the simulator id will be
        looked up, but it will take a bit longer.""")
    parser.add_argument(
        '--sleep', '-s', metavar='<sleep-seconds>', default=600, type=int,
        help="""Amount of time to sleep in seconds while waiting for output
        from the simulator. Too long and the script will be paused waiting for
        a response from the simulator, too short and a lof of cpu cycles will
        be wasted checking an empty buffer. (default: %(default)d)""")
    parser.add_argument(
        '--max-schedule', metavar='<observations>', default=100, type=int,
        help="""The maximum number of observations to schedule simultaneously.
        This isn't a strict limit, but it won't be violated by much. (default:
        %(default)s)""")
    parser.add_argument(
        '--sim-memory', '-m', required=True, metavar='<mb>', type=int,
        help="""Maximum memory needed to run simulation. Standard limit per
        core is 4096, but if your simulation needs less, you should request
        less. Note, that if you set it too low you'll get biased samples for
        simulations that require less memory to run which could be very
        bad.""")
    parser.add_argument(
        '--sim-time', '-t', required=True, metavar='<seconds>', type=int,
        help="""Maximum amount of time needed to run a single simulation.
        Longer times will result in simulations taking longer to schedule, but
        shorter times could result in jobs being canceled from going over time
        which will bias payoff results for shorter running simulations.""")
    return parser


def create_scheduler(game, serial, args, configuration=None, simname=None,
                     **_):
    assert configuration is not None or args.conf is not None, \
        "`conf` must be specified or supplied in the game"
    assert simname is not None or args.sim_id is not None, \
        "`sim-id` must be specified or supplied in the game"
    if args.conf is not None:
        configuration = json.load(args.conf)

    if args.sim_id is None:
        with api.EgtaOnlineApi(auth_token=args.auth_string) as ea:
            args.sim_id = next((
                s['id'] for s in ea.get_simulators()
                if '{}-{}'.format(s['name'], s['version']) == simname), None)
    if args.sim_id is None:  # pragma: no cover
        _log.critical('couldn\'t find simulator with name "{}"'.format(
            simname))
        sys.exit(1)

    return egtasched.EgtaScheduler(
        args.sim_id, game, serial, args.count, configuration, args.sleep,
        args.max_schedule, args.sim_memory, args.sim_time, args.auth_string)



"""Python script for performing egta"""
import argparse
import json
import logging
import smtplib
import sys
import traceback
from logging import handlers

from egtaonline import api
from gameanalysis import gameio

from egta import countsched
from egta import savesched
from egta.script import brute
from egta.script import egtasched
from egta.script import gamesched
from egta.script import outerloop
from egta.script import simsched


_log = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        description="""Command line egta. To run, both an equilibrium finding
        method and a profile scheduler must be specified. Each element
        processes the arguments after it, e.g. `egta -a brute -b game -c` will
        pass -a to the base parser, -b to the brute equilibrium solver, and -c
        to the game profile scheduler.""")

    # Standard arguments
    parser.add_argument(
        '--output', '-o', metavar='<output-file>', type=argparse.FileType('w'),
        default=sys.stdout, help="""The file to write the output to. (default:
        stdout)""")
    parser.add_argument(
        '--verbose', '-v', action='count', default=0, help="""Sets the
        verbosity of commands. Output is send to standard error.""")
    parser.add_argument(
        '-e', '--email_verbosity', action='count', default=0, help="""Verbosity
        level for emails.""")
    parser.add_argument(
        '-r', '--recipient', metavar='<email-address>', action='append',
        default=[], help="""Specify an email address to receive email logs at.
        Can specify multiple email addresses.""")
    parser.add_argument(
        '--count', metavar='<sims>', type=int, default=1, help="""Number of
        simulations to schedule to count towards a single payoff. (default:
        %(default)d)""")
    parser.add_argument(
        '--profile-data', metavar='<gamefile>', help="""If specified, record
        all of the payoff data sampled, and write them as an observation game
        file to the specified location.""")
    parser.add_argument(
        '--support-thresh', metavar='<support-thresh>', type=float,
        default=1e-3, help="""Threshold for a strategy probability to consider
        it in support. (default: %(default)s)""")
    parser.set_defaults(loaders=(lambda g: g.get_summary(),
                                 gameio.read_basegame))

    # Egta online authentication
    parser_auth = parser.add_mutually_exclusive_group()
    parser_auth.add_argument(
        '--auth-string', '-a', metavar='<auth-string>', help="""The string
        authorization token to connect to egta online.""")
    parser_auth.add_argument(
        '--auth-file', '-f', metavar='<auth-file>', help="""Filename that just
        contains the string of the auth token.""")

    # Information about the game to sample
    game_spec = parser.add_mutually_exclusive_group(required=True)
    game_spec.add_argument(
        '--game-id', '-g', metavar='<game-id>', type=int, help="""An egtaonline
        game id that specifies that game profile to load. A game profile is a
        configuration, number of players per role, and strategy names per
        roll.""")
    game_spec.add_argument(
        '--game-json', metavar='<game-json>', type=argparse.FileType('r'),
        help="""A game file that must specify the number of players per role,
        and names of all role and strategies in a standard game analysis
        format.""")

    # All of the actual methods to run
    eq_methods = parser.add_subparsers(
        title='methods', dest='method', metavar='<methods>', help="""The
        method to use to compute equilibria. Available commands are:""")
    for module in [outerloop, brute]:
        method = module.add_parser(eq_methods)
        method.find_eqa = module.find_eqa

        schedulers = method.add_subparsers(
            title='schedulers', dest='scheduler', metavar='<scheduler>',
            help="""The scheduler to use for acquiring samples from profiles.
            Available commands are:""")
        method.sched = schedulers
        for module in [egtasched, gamesched, simsched]:
            sched = module.add_parser(schedulers)
            sched.create_scheduler = module.create_scheduler

    # Parse args and process
    args = parser.parse_args()
    method = eq_methods.choices[args.method]
    sched = method.sched.choices[args.scheduler]

    if args.auth_string is None and args.auth_file is not None:
        with open(args.auth_file) as auth_file:  # pragma: no cover
            args.auth_string = auth_file.read().strip()

    if args.game_id is not None:  # pragma: no cover
        fmt_str = '%(asctime)s ({gid:d}) %(levelname)s %(message)s'.format(
            gid=args.game_id)
        email_subject = 'EGTA Online Quiesce Status for Game {gid:d}'.format(
            gid=args.game_id)
    else:
        fmt_str = '%(asctime)s %(levelname)s %(message)s'
        email_subject = 'EGTA Online Quiesce Status'

    stderr_handle = logging.StreamHandler(sys.stderr)
    stderr_handle.setLevel(50 - 10 * min(args.verbose, 4))
    stderr_handle.setFormatter(logging.Formatter(fmt_str))
    log_handlers = [stderr_handle]

    # Email Logging
    if args.recipient:  # pragma: no cover
        smtp_host = 'localhost'

        # We need to do this to match the from address to the local host name
        # otherwise, email logging will not work. This seems to vary somewhat
        # by machine
        with smtplib.SMTP(smtp_host) as server:
            smtp_fromaddr = 'EGTA Online <egta_online@{host}>'.format(
                host=server.local_hostname)

        email_handler = handlers.SMTPHandler(smtp_host, smtp_fromaddr,
                                             args.recipient, email_subject)
        email_handler.setLevel(40 - args.email_verbosity * 10)
        log_handlers.append(email_handler)

    logging.basicConfig(level=0, handlers=log_handlers)

    try:
        if args.game_id is not None:
            with api.EgtaOnlineApi(auth_token=args.auth_string) as ea:
                jgame = args.loaders[0](ea.get_game(args.game_id))
        else:
            jgame = json.load(args.game_json)
        game, serial = args.loaders[1](jgame)

        extra = {}
        if 'configuration' in jgame:
            extra['configuration'] = dict(jgame['configuration'])
        if 'simulator_fullname' in jgame:
            extra['simname'] = jgame['simulator_fullname']

        sched = sched.create_scheduler(game, serial, args, **extra)
        if args.profile_data is not None:
            prof_data = savesched.SaveScheduler(game, sched)
            sched = prof_data
        sched = countsched.CountScheduler(sched, args.count)
        with sched:
            eqa = method.find_eqa(sched, game, serial, args)

        if args.profile_data is not None:
            gamej = serial.to_samplegame_json(prof_data.get_samplegame())
            with open(args.profile_data, 'w') as f:
                json.dump(gamej, f)

        for eqm in eqa:
            json.dump(
                serial.to_prof_json(game.trim_mixture_support(
                    eqm, args.support_thresh)), args.output)
            args.output.write('\n')
    except KeyboardInterrupt:  # pragma: no cover
        _log.critical('execution interrupted by user or exception in thread')

    except Exception as ex:  # pragma: no cover
        exc_type, exc_value, exc_traceback = sys.exc_info()
        _log.critical(''.join(traceback.format_exception(
            exc_type, exc_value, exc_traceback)))
        raise ex

# Empty line to stop vim from interpreting last except as a modeline



import argparse

from gameanalysis import gameio
from gameanalysis import rsgame

from egta import gamesched


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'game', help="""Schedule simulations with noise added to a game""",
        description="""Sample profiles by adding noise to game data.""")

    load_type = parser.add_mutually_exclusive_group(required=True)
    load_type.add_argument(
        '--load-game', dest='loaders', action='store_const',
        const=(lambda g: g.get_summary(), gameio.read_game),
        default=argparse.SUPPRESS, help="""Load game with data instead of just
        format.""")
    load_type.add_argument(
        '--load-samplegame', dest='loaders', action='store_const',
        const=(lambda g: g.get_observations(), gameio.read_samplegame),
        default=argparse.SUPPRESS, help="""Load game with sample data instead
        of just format.""")
    return parser


def create_scheduler(game, serial, args, **_):
    if isinstance(game, rsgame.SampleGame):
        return gamesched.SampleGameScheduler(game)
    else:
        return gamesched.GameScheduler(game)



import setuptools


setuptools.setup(
    name='egta',
    version='0.0.1',
    description='Scripts to perform EGTA',
    url='https://github.com/egtaonline/quiesce.git',
    author='Strategic Reasoning Group',
    author_email='strategic.reasoning.group@umich.edu',
    license='Apache 2.0',
    entry_points=dict(console_scripts=['egta=egta.egta:main']),
    install_requires=[
        'gameanalysis~=3.1',
        'egtaonlineapi~=0.1',
        'numpy~=1.12',
        'tabulate~=0.7',
    ],
    packages=['egta', 'egta.script'],
)



"""A scheduler that gets payoffs from a local simulation"""
import io
import json
import logging
import queue
import subprocess
import sys
import threading
import time
import traceback

import numpy as np

from egta import profsched


_log = logging.getLogger(__name__)


class SimulationScheduler(profsched.Scheduler):
    """Schedule profiles using a command line program

    Parameters
    ----------
    serial : GameSerializer
        A gameanalysis game serializer that indicates how array profiles should
        be turned into json profiles.
    config : {key: value}
        A dictionary mapping string keys to values that will be passed to the
        simulator in the standard simulation spec format.
    command : [str]
        A list of strings that represents a command line program to run. This
        program must accept simulation spec files as flushed lines of input to
        standard in, and write the resulting output as an observation to
        standard out. After all input lines have been read, this must flush the
        output otherwise this could hang waiting for results that are trapped
        in a buffer.
    sleep : int, optional
        Time in seconds to wait before checking programs stdout for results.
        Too low and a lot of cycles will be wasted querying an empty buffer,
        too fast and programs may be waiting for results while this is
        sleeping.
    """

    def __init__(self, serial, config, command, sleep=1):
        self.serial = serial
        self.base = {'configuration': config}
        self.command = command
        self.sleep = sleep

        self._running = False
        self._proc = None
        self._stdin = None
        self._stdout = None
        self._lock = threading.Lock()
        self._queue = queue.Queue()
        self._thread_failed = False

    def schedule(self, profile):
        promise = _SimPromise(self)
        with self._lock:
            self.base['assignment'] = self.serial.to_prof_json(profile)
            json.dump(self.base, self._stdin, separators=(',', ':'))
            self._stdin.write('\n')
            self._stdin.flush()
            self._queue.put(promise)
            _log.debug("sent profile: %s", profile)
        return promise

    def _dequeue(self):
        try:
            while self._running:
                line = self._stdout.readline()
                if not line:
                    time.sleep(self.sleep)
                else:
                    payoffs = self.serial.from_payoff_json(json.loads(line))
                    _log.debug("read payoff: %s", payoffs)
                    promise = self._queue.get()
                    promise._set(payoffs)
                    self._queue.task_done()
        except Exception as ex:  # pragma: no cover
            self._thread_failed = True
            exc_type, exc_value, exc_traceback = sys.exc_info()
            _log.critical(''.join(traceback.format_exception(
                exc_type, exc_value, exc_traceback)))
            while not self._queue.empty():
                self._queue.get()._set(None)  # Trigger check for failure
            raise ex

    def __enter__(self):
        self._proc = subprocess.Popen(
            self.command, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
        self._stdin = io.TextIOWrapper(self._proc.stdin)
        self._stdout = io.TextIOWrapper(self._proc.stdout)

        self._running = True
        self._thread_failed = False
        threading.Thread(target=self._dequeue, daemon=True).start()

        return self

    def __exit__(self, *args):
        self._running = False
        if self._proc is not None:
            self._proc.kill()


class _SimPromise(profsched.Promise):
    def __init__(self, sched):
        self._event = threading.Event()
        self._sched = sched

    def _set(self, value):
        self._value = value
        self._event.set()

    def get(self):
        self._event.wait()
        if self._sched._thread_failed:
            raise RuntimeError("a thread failed, checl log for details")  # pragma: no cover # noqa
        else:
            return self._value
